import random

# an alignment like
# TCA--CCCAT
# GGAGGTTGGT
# is represented in a list as ['TCA--CCCAT','GGAGGTTGGT']



# a similarity matrix is represented by a dict of dicts, so that the similarity
# of base T in seq 1 and G in seq 2 can be referred as, e.g., sim['T']['G']

# some possible similarity matrices are reported hereafter

sim_m_0 = {'A': {'A':10, 'G':-1, 'C':-3, 'T':-4}, \
    'G': {'A':-1, 'G':7, 'C':-5, 'T':-3}, \
    'C': {'A':-3, 'G':-5, 'C':9, 'T':0}, \
    'T': {'A':-4, 'G':-3, 'C':0, 'T':8}}

sim_m_1 = {'A': {'A':10, 'G':-6, 'C':-6, 'T':-6}, \
    'G': {'A':-6, 'G':10, 'C':-6, 'T':-6}, \
    'C': {'A':-6, 'G':-6, 'C':10, 'T':-6}, \
    'T': {'A':-6, 'G':-6, 'C':-6, 'T':10}}

#the next is the one used in BLAST
sim_BLAST = {'A': {'A':5, 'G':-4, 'C':-4, 'T':-4}, \
    'G': {'A':-4, 'G':5, 'C':-4, 'T':-4}, \
    'C': {'A':-4, 'G':-4, 'C':5, 'T':-4}, \
    'T': {'A':-4, 'G':-4, 'C':-4, 'T':5}}

#the next one considers transitions/transvections
sim_TR_TR = {'A': {'A':1, 'G':-1, 'C':-5, 'T':-5}, \
    'G': {'A':-1, 'G':1, 'C':-5, 'T':-5}, \
    'C': {'A':-5, 'G':-5, 'C':1, 'T':-1}, \
    'T': {'A':-5, 'G':-5, 'C':-1, 'T':1}}

#the next one corresponds to -1*(Levenshtein) (-edit_distance)
sim_neg_Leven = {'A': {'A':0, 'G':-1, 'C':-1, 'T':-1}, \
    'G': {'A':-1, 'G':0, 'C':-1, 'T':-1}, \
    'C': {'A':-1, 'G':-1, 'C':0, 'T':-1}, \
    'T': {'A':-1, 'G':-1, 'C':-1, 'T':0}}



def score_alignment(al, simil_matrix, gp):
    """Returns the score of alignment 'al', 
       given simil. matrix & gap penalty"""
    score = 0
    for ca, cb in zip(al[0], al[1]):
        if ca != '-' and cb != '-':
            score += simil_matrix[ca][cb]
        else: score += gp
    return score



def compute_F_B(a, b, similarity_matrix, gap_penalty):
    """
    Returns the score matrix F for two sequences a and b to be aligned.
    F is the matrix for partial scores.
    The matrix is represented as a list of lists, with #columns = len(a)+1
    and #rows = len(b)+1
    Moreover, it returns also the matrix B with the possible backsteps.
    """

    len_a, len_b = len(a), len(b)
    aa, bb = a.upper(), b.upper()
    
    #initialization of F    
    matrix_F = [[gap_penalty * j for j in xrange(len_a + 1)]]
    for i in xrange(1, len_b + 1):
        matrix_F.append([gap_penalty*i])
    #initialization of B    
    matrix_B = [[[0]]]
    matrix_B[0].extend([[1] for j in xrange(len_a)])
    for i in xrange(1, len_b + 1):
        matrix_B.append([[3]])

    #computation of scores and backsteps
    for i in xrange(1, len_b + 1):
        for j in xrange(1, len_a + 1):
            val1 = matrix_F[i][j-1] + gap_penalty
            val2 = matrix_F[i-1][j-1] + similarity_matrix[aa[j-1]][bb[i-1]]
            val3 = matrix_F[i-1][j] + gap_penalty
            val = max(val1, val2, val3)
            matrix_F[i].append(val)
            
            moves=[]
            if val1 == val:
                moves.append(1) # 1 means leftwards 
            if val2 == val:
                moves.append(2) # 3 means left-upwards (diagonal)
            if val3 == val:
                moves.append(3) # 3 means upwards
            matrix_B[i].append(moves)
            
    return matrix_F, matrix_B




def galign_simple(a, b, simil_matrix, gap_penalty):
    """
    Returns only one global alignments with max score.
    Such an alignment corresponds to the leftmost choice
    in the backsteps matrix.
    """

    matrix_F, matrix_B = compute_F_B(a,b,simil_matrix, gap_penalty)

    al_upper_seq = al_lower_seq = ''
    i, j = len(b), len(a)
    step = 4 #just to start the backtrace

    while step > 0:
        step = matrix_B[i][j][0]
        if step == 1:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = '-' + al_lower_seq
            j -= 1
        elif step == 2:
            al_upper_seq = a[j-1] + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1
            j -= 1
        elif step == 3:
            al_upper_seq = '-' + al_upper_seq
            al_lower_seq = b[i-1] + al_lower_seq
            i -= 1
        
    return [al_upper_seq, al_lower_seq]



def galign_list(a, b, simil_matrix, gap_penalty):
    """
    Returns a list with all the global alignments with max score
    """

    matrix_F, matrix_B = compute_F_B(a, b, simil_matrix, gap_penalty)
    all_alignments = []

    def get_align_list(j, i, partial_a, partial_b):
                    
        steps = matrix_B[i][j]
        for step in steps:
            if step == 1:
                new_partial_a = a[j-1] + partial_a
                new_partial_b  = '-' + partial_b
                get_align_list(j-1, i, new_partial_a[:], new_partial_b[:])
            elif step == 2:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = b[i-1] + partial_b
                get_align_list(j-1, i-1, new_partial_a[:], new_partial_b[:])
            elif step == 3:
                new_partial_a = '-' + partial_a
                new_partial_b = b[i-1] + partial_b
                get_align_list(j, i-1, new_partial_a[:], new_partial_b[:])
            elif step == 0:
                current_alignment = [partial_a, partial_b]
                all_alignments.append(current_alignment)

        return
   
    get_align_list(len(matrix_F[0])-1, len(matrix_F)-1, '', '')

    return all_alignments



class Node(object):

    def __init__(self, info, childrenlist = None):
        self.info = info
        if childrenlist:
            self.clist = childrenlist
        else:
            self.clist = []

    def addChild(self, child):
        self.clist.append(child)
    
    def __str__(self):
        res = ""
        if self.clist:
            for child in self.clist:
                res += str(child)
        else:
            res = str(self.info) + "\n\n"
        return res
        


def galign_tree(a, b, simil_matrix, gap_penalty):
    """
    global alignment with Needleman Wunsch;
    returns a tree with all the partial alignments that lead 
    to the global alignments with max score (in the tree leaves)
    """

    matrix_F, matrix_B = compute_F_B(a, b, simil_matrix, gap_penalty)

    def get_align_tree(j, i, partial_a, partial_b):
        
        currentNode = Node([partial_a,partial_b])
            
        for step in matrix_B[i][j]:
            if step == 1:
                new_partial_a = a[j-1] + partial_a
                new_partial_b  = '-' + partial_b
                cld = get_align_tree(j-1, i, new_partial_a, new_partial_b)
            elif step == 2:
                new_partial_a = a[j-1] + partial_a
                new_partial_b = b[i-1] + partial_b
                cld = get_align_tree(j-1, i-1, new_partial_a, new_partial_b)
            elif step == 3:
                new_partial_a = '-' + partial_a
                new_partial_b = b[i-1] + partial_b
                cld = get_align_tree(j, i-1, new_partial_a, new_partial_b)
            elif step == 0:
                break
            currentNode.addChild(cld)

        return currentNode
   
    return get_align_tree(len(a), len(b), '', '')



def rnd_DNA_seq(length):
    """Returns a random sequence of nucleotides, whose length is 'len'"""
    return "".join(random.choice('ATCG') for i in xrange(length))
